#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Thread-safe connection cache
#
# Copyright (c) 2006 JAS
#
# Author: Petr Vokac <vokac@kmlinux.fjfi.cvut.cz>
#
# $Id$
#
import logging
import threading



class ConnCache:

    lock = threading.Lock()
    cache = {}


    def add(self, key, client):
        logging.getLogger().debug("%s.add(%s): %s" % (self.__class__.__name__, client, key))
#
#        if client.peer == None:
#            logging.getLogger().debug("peer is dead")
#            return

        self.lock.acquire()
        try:
            self.cache.setdefault(key, []).append(client)
        finally:
            self.lock.release()


    def remove(self, key, client = None):
        logging.getLogger().debug("%s.remove(%s): %s" % (self.__class__.__name__, client, key))
        retVal = []

        self.lock.acquire()
        try:
            if self.cache.has_key(key):
                if client == None:
                    retVal = self.cache[key]
                    del self.cache[key]
                else:
                    newConns = []
                    for conn in self.cache[key]:
                        if conn != client:
                            newConns.append(conn)
                        else:
                            retVal.append(conn)
                    if len(newConns) == 0:
                        del self.cache[key]
                    else:
                        self.cache[key] = newConns
        finally:
            self.lock.release()

        for conn in retVal:
            conn.transport.loseConnection()
#        return retVal


    def removeAll(self, host = None, port = None, tls_use = None, user = None, password = None):
        logging.getLogger().debug("%s.removeAll(%s, %s, %s, %s, %s)" % (self.__class__.__name__, host, port, tls_use, user, password))
        retVal = []

        self.lock.acquire()
        try:
            for key, clients in self.cache:
                if host == None or key[0].lower() != host.lower(): continue
                if port == None or key[1].lower() != port.lower(): continue
                if tls_use == None or key[2].lower() != tls_use.lower(): continue
                if user == None or key[3].lower() != user.lower(): continue
                if password == None or key[4].lower() != password.lower(): continue
                retVal.append(key)
            for key in retVal:
                del self.cache[key]
        finally:
            self.lock.release()

        for conn in retVal:
            conn.transport.loseConnection()
#        return retVal
        


    def get(self, key):
        logging.getLogger().debug("%s.get(%s)" % (self.__class__.__name__, key))
        retVal = None

        self.lock.acquire()
        try:
            if self.cache.has_key(key):
                retVal = self.cache[key].pop()
                if len(self.cache[key]) == 0:
                    del self.cache[key]
                # FIXME: check if connection is still active
        finally:
            self.lock.release()

        return retVal
