#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Horde IMAP Proxy server for CTU FNSPE
#
# Copyright (c) 2006 JAS
#
# Author: Petr Vokac <vokac@kmlinux.fjfi.cvut.cz>
#
# $Id$
#
import logging
import ssl
from twisted.protocols.portforward import Proxy
#from twisted.protocols.portforward import ProxyFactory
from twisted.internet import reactor, protocol
import twisted.internet.ssl



class ImapPop3ProxyClient(Proxy):

    auth = False
    user = None
    password = None

    def __init__(self, server, connCache):
        logging.getLogger().debug("%s.__init__" % self.__class__.__name__)
        self.peer = server
        self.connCache = connCache
        self.host = self.peer.config.get('host')
        self.port = self.peer.config.get('port')
        self.tls_use = self.peer.config.get('tls_use', False)
        self.user = self.peer.user
        self.password = self.peer.password

#    def setServer(self, server):
#        self.peer = server
#        #self.config = server.config
#        #self.connCache = connCache

    def connectionMade(self):
        logging.getLogger().debug("%s.connectionMade" % self.__class__.__name__)
        self.peer.setPeer(self)

    def connectionLost(self, reason):
        logging.getLogger().debug("%s.connectionLost(%s)" % (self.__class__.__name__, reason))
        key = (self.host, self.port, self.tls_use, self.user, self.password)
        self.connCache.remove(key, self)
        if self.peer != None:
            if self.peer.peer != None:
                self.peer.peer = None
            self.peer.transport.loseConnection()
            del self.peer

    def dataReceived(self, data):
        logging.getLogger().debug("%s.dataReceived(): %s/%s/%s server send: %s" % (self.__class__.__name__, self.host, self.port, self.user, repr(data)))
        Proxy.dataReceived(self, data)



class ImapPop3ProxyClientFactory(protocol.ClientFactory):

    protocol = ImapPop3ProxyClient

    def __init__(self, server):
        logging.getLogger().debug("%s.__init__" % self.__class__.__name__)
        self.server = server

    def buildProtocol(self, addr):
        logging.getLogger().debug("buildProtocol: %s" % addr)
        return self.protocol(self.server, self.server.connCache)

#    def clientConnectionFailed(self, connector, reason):
#        if self.peer != None:
#            self.peer.transport.loseConnection()



class ImapPop3ProxyServer(Proxy):

    clientProtocolFactory = ImapPop3ProxyClientFactory
    user = None
    password = None

    def __init__(self, config, connCache):
        logging.getLogger().debug("%s: %s" % (self.__class__.__name__, config))
        self.config = config
        self.connCache = connCache

    def connectionLost(self, reason):
        logging.getLogger().debug("%s.connectionLost(%s): %s" % (self.__class__.__name__, reason, self.peer))
        if self.peer != None:
            self.peer.peer = None
            key = (self.peer.host, self.peer.port, self.peer.tls_use, self.peer.user, self.peer.password)
            self.connCache.add(key, self.peer)

    def dataReceived(self, data):
        logging.getLogger().debug("%s.dataReceived(): %s/%s/%s client send: %s" % (self.__class__.__name__, self.config.get('host'), self.config.get('port'), self.user, repr(data)))
        Proxy.dataReceived(self, data)
        

    def getClientConnection(self, user, password):
        logging.getLogger().debug("%s.getClientConnection(%s, %s)" % (self.__class__.__name__, user, password))
        self.peer = self.connCache.get((self.config.get('host'),
                                        self.config.get('port'),
                                        self.config.get('tls_use', False),
                                        user, password))
        if self.peer != None:
            logging.getLogger().debug("cached auth connection for: %s" % self.peer)
            self.peer.setPeer(self)
            return True

        self.peer = self.connCache.get((self.config.get('host'),
                                        self.config.get('port'),
                                        self.config.get('tls_use', False),
                                        None, None))
        if self.peer != None:
            logging.getLogger().debug("cached unauth connection: %s" % self.peer)
            self.peer.setPeer(self)
            return False

        clientFactory = self.clientProtocolFactory(self)

        if not self.config.get('tls_use', False):
            reactor.connectTCP(self.config.get('host'),
                               self.config.get('port'),
                               clientFactory)
        else:
            if self.config.get('tls_checkpeer', False):
                ssl_context = ssl.VerificationClientContextFactory(self.config.get('tls_cacertfile'), self.config.get('tls_cacertdir'))
            else:
                ssl_context = twisted.internet.ssl.ClientContextFactory()
            reactor.connectSSL(self.config.get('host'),
                               self.config.get('port'),
                               clientFactory, ssl_context)

        logging.getLogger().debug("created new connection: %s" % self.peer)
        return False



class Pop3ProxyServer(ImapPop3ProxyServer):

    def connectionMade(self):
        self.transport.write("+OK POP3 Proxy server ready.\r\n")

    def dataReceived(self, data):
        logging.getLogger().debug("%s.dataReceived(): %s/%s/%s send: %s" % (self.__class__.__name__, self.config.get('host'), self.config.get('port'), self.user, repr(data)))
        if self.peer != None:
            data = "%s " % data.rstrip()
            if data[:5].upper() == 'QUIT ':
                # don't logout keep connection till timeout
                self.transport.write("+OK Logging out from POP3 Proxy - fake.\r\n")
                self.transport.loseConnection()
                return
            ImapPop3ProxyServer.dataReceived(self, data)
        else:
            cmdline = "%s " % data.rstrip()
            cmd, cmdpar = cmdline.split(" ", 1)
            cmd = cmd.upper()
            if cmd == 'USER':
                self.user = cmdpar
                self.transport.write("+OK\r\n")
            elif cmd == 'PASS':
                self.password = cmdpar

                if self.user != None:
                    if not self.getClientConnection(self.user, self.password):
#                        self.peer.discardResponse(True)
                        data = "USER %s\r\n" % self.user
                        ImapPop3ProxyServer.dataReceived(self, data)
                        data = "PASS %s\r\n" % self.password
                        ImapPop3ProxyServer.dataReceived(self, data)
                    else:
                        self.transport.write("+OK\r\n")
            elif cmd == 'QUIT':
                self.transport.write("DONE\r\n")
                self.transport.loseConnection()
            else:
                self.transport.write("-ERR Unknown command.\r\n")



class ImapProxyServer(ImapPop3ProxyServer):

    def connectionMade(self):
        self.transport.write("* OK IMAP Proxy server ready.\r\n")

    def dataReceived(self, data):
        logging.getLogger().debug("%s.dataReceived(): %s/%s/%s send: %s" % (self.__class__.__name__, self.config.get('host'), self.config.get('port'), self.user, repr(data)))
        if self.peer != None:
            if data.find(' ') != -1:
                tag, cmdline = data.split(' ', 1)
                cmdline = "%s " % cmdline.rstrip()
                if tag != '' and cmdline[:7].upper() == 'LOGOUT ':
                    # don't logout keep connection till timeout
                    self.transport.write("* BYE Logging out from IMAP Proxy - fake\r\n")
                    self.transport.write("%s OK LOGOUT completed from IMAP Proxy - fake.\r\n" % tag)
                    self.transport.loseConnection()
                    return
            ImapPop3ProxyServer.dataReceived(self, data)
        else:
            if data.find(' ') == -1:
                if data.rstrip() == '':
                    tag = '*'
                else:
                    tag = data
                self.transport.write("%s BAD Protocol Error: \"No space following tag in IMAP command\".\r\n" % tag)
                return
            tag, cmdline = data.split(' ', 1)
            cmdline = cmdline.rstrip()
            if tag == '':
                self.transport.write("* BAD Protocol Error: \"Tag not found in command\".\r\n")
                return
            if cmdline.find(' ') == -1:
                cmd = cmdline.upper()
                cmdpar = ''
            else:
                cmd, cmdpar = cmdline.split(' ', 1)
                cmd = cmd.upper()
            if cmd == 'LOGIN':
                # parse login params
                if cmdpar.find(' ') == -1:
                    self.user = cmdpar
                    self.password = ''
                else:
                    self.user, self.password = cmdpar.split(' ', 1)

                if not self.getClientConnection(self.user, self.password):
                    ImapPop3ProxyServer.dataReceived(self, data)
                else:
                    self.transport.write("%s OK Logged in - cached.\r\n" % tag)
            elif cmd == 'NOOP':
                self.transport.write("%s OK %s completed.\r\n" % (tag, cmd))
            elif cmd == 'CAPABILITY':
                # FIXME: what to return?
# Dovecot:      * CAPABILITY IMAP4rev1 SORT THREAD=REFERENCES MULTIAPPEND UNSELECT LITERAL+ IDLE CHILDREN LISTEXT LIST-SUBSCRIBED NAMESPACE AUTH=PLAIN
# Exchange:     * CAPABILITY IMAP4 IMAP4rev1 IDLE LOGIN-REFERRALS MAILBOX-REFERRALS NAMESPACE LITERAL+ UIDPLUS CHILDREN AUTH=NTLM
# Sun Java SMS: * CAPABILITY IMAP4 IMAP4rev1 ACL QUOTA LITERAL+ NAMESPACE UIDPLUS CHILDREN BINARY LANGUAGE STARTTLS XSENDER X-NETSCAPE XSERVERINFO AUTH=PLAIN
# uw-imap?:     * CAPABILITY IMAP4REV1 LITERAL+ IDLE NAMESPACE MAILBOX-REFERRALS BINARY UNSELECT SCAN SORT THREAD=REFERENCES THREAD=ORDEREDSUBJECT MULTIAPPEND SASL-IR LOGIN-REFERRALS AUTH=PLAIN AUTH=LOGIN
# Mercury:      * CAPABILITY IMAP4rev1 STARTTLS X-MERCURY
#                self.transport.write("* CAPABILITY IMAP4rev1 IDLE NAMESPACE AUTH=PLAIN\r\n")
                self.transport.write("* CAPABILITY IMAP4rev1\r\n")
                self.transport.write("%s OK %s completed.\r\n" % (tag, cmd))
            elif cmd == 'AUTHENTICATE': # FIXME: SASL AUTH
                self.transport.write("%s NO Unsupported authentication mechanism.\r\n" % tag)
            elif cmd == 'LOGOUT':
                self.transport.write("* BYE Logging out from IMAP Proxy\r\n")
                self.transport.write("%s OK %s completed.\r\n" % (tag, cmd))
                self.transport.loseConnection()
            else:
                self.transport.write("%s BAD Error in IMAP command %s: Unknown command.\r\n" % (tag, cmd))


         
class ImapPop3ProxyFactory(protocol.Factory):
    """Factory for port forwarder."""

    protocol = None

    def __init__(self, protocol, config, connCache):
        self.protocol = protocol
        self.config = config
        self.connCache = connCache

    def buildProtocol(self, addr):
        logging.getLogger().debug("buildProtocol: %s" % addr)
        return self.protocol(self.config, self.connCache)
