#!/usr/bin/python
# -*- coding: utf-8 -*-
supported = False

# System imports
from OpenSSL import SSL
import logging
import os, os.path

# Twisted imports
from twisted.internet import ssl


class VerificationClientContextFactory(ssl.ClientContextFactory):
    """A context factory for SSL clients."""

    isClient = 1
    method = SSL.SSLv3_METHOD

    def __init__(self, tlsCACertFile = None, tlsCACertDir = None,
                 sslmethod=SSL.SSLv23_METHOD):
        logging.getLogger().debug("%s.__init__(%s, %s, %s)" % (self.__class__.__name__, tlsCACertFile, tlsCACertDir, sslmethod))
        self.tlsCACertFile = tlsCACertFile
        self.tlsCACertDir = tlsCACertDir
        self.sslmethod = sslmethod
        self.cacheContext()

    def cacheContext(self):
        ctx = SSL.Context(self.sslmethod)
        # ctx.set_options(SSL.OP_NO_SSLv2)
        ctx.set_verify(SSL.VERIFY_PEER, lambda conn, cert, errnum, depth, ok: ok)
#        ctx.use_certificate_file(self.certificateFileName)
#        ctx.use_privatekey_file(self.privateKeyFileName)
        if self.tlsCACertFile != None:
            logging.getLogger().debug("CA file: %s" % self.tlsCACertFile)
            ctx.load_verify_locations(self.tlsCACertFile)
        if self.tlsCACertDir != None:
            for dirobj in os.listdir(self.tlsCACertDir):
                if os.path.isfile(os.path.join(self.tlsCACertDir, dirobj)):
                    if dirobj[-2:] == '.0':
                        logging.getLogger().debug("CA file: %s" % os.path.join(self.tlsCACertDir, dirobj))
                        ctx.load_verify_locations(os.path.join(self.tlsCACertDir, dirobj))
        self._context = ctx

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_context']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self.cacheContext()

    def getContext(self):
        """Create an SSL context.
        """
        return self._context


__all__ = ["VerificationClientContextFactory"]

supported = True
